# pylint: disable=super-with-arguments
# pylint: disable=redefined-builtin
# pylint: disable=invalid-name

"""
    ATHENA TASK CODE
        - these are used in pythonOperator steps as a way \
            to flexibly call boto3 and perform any special requirements
"""
import os
import logging
from airflow.models import BaseOperator
from airflow.utils.decorators import apply_defaults
import boto3
from botocore.exceptions import ClientError as botocore_clienterror

logger = logging.getLogger(None)

class StartQueryExecution(BaseOperator):
    template_fields = ('config',)

    @apply_defaults
    def __init__(self, config, *args, **kwargs):
        super(StartQueryExecution, self).__init__(*args, **kwargs)
        self.config = config

    def execute(self, context):
        """
            Runs athena query against feature store

            https://sagemaker.readthedocs.io/en/stable/api/\
                prep_data/feature_store.html#sagemaker.feature_store.feature_group.AthenaQuery
        """
        # init boto session and sagemaker
        BOTO_SESSION = boto3.Session()
        ATHENA_CLIENT = BOTO_SESSION.client('athena')

        try:
            response = ATHENA_CLIENT.start_query_execution(
                QueryString=self.config['query'],
                QueryExecutionContext=self.config.get('query_execution_context', {}),
                ResultConfiguration=self.config['result_configuration'],    
                WorkGroup=self.config['workgroup']
            )

            context['task_instance'].xcom_push(
                key='athena_query_execution_id',
                value=response['QueryExecutionId']
            )

        except Exception as err:
            raise



"""Custom Sagemaker Operators"""
import logging
from airflow.models import BaseOperator
from airflow.utils.decorators import apply_defaults

import boto3
from sagemaker.session import Session
from sagemaker.estimator import Estimator, Transformer

logger = logging.getLogger(None)


class CreateTrainingJob(BaseOperator):
    """Execute Create Sagemaker training job using boto3 api"""
    template_fields = ('config', 'xcom_tasks')

    @apply_defaults
    def __init__(self, config, xcom_tasks=None, *args, **kwargs):
        super(CreateTrainingJob, self).__init__(*args, **kwargs)
        self.config = config
        self.xcom_tasks = xcom_tasks

    def execute(self, context):
        """
            Execute a sagemaker training job run
        """
        # init boto session and sagemaker
        BOTO_SESSION = boto3.Session()
        SM_SESSION = Session(default_bucket=self.config.pop('bucket', None), boto_session=BOTO_SESSION)

        print("Running else where")
        IAM_CLIENT = BOTO_SESSION.client('iam')
        ROLE = IAM_CLIENT.get_role(RoleName=self.config.pop('sagemaker_role_name', {}))['Role']['Arn']

        input_data = self.config.pop('inputs', None)
        wait_for_completion = self.config.pop('wait', False)

        try:
            print("creating estimator")
            estimator = Estimator(
                role=ROLE,
                sagemaker_session=SM_SESSION,
                **self.config
            )

            print("starting training")
            estimator.fit(
                inputs = input_data,
                wait=wait_for_completion
            )

            context['task_instance'].xcom_push(
                key='training_job_name',
                value=estimator.latest_training_job.name
            )
            context['task_instance'].xcom_push(
                key='model_name',
                value=estimator.latest_training_job.name
            )
        except Exception as exception:
            logger.error(exception)
            raise

class CreateModel(BaseOperator):
    """Execute Create Sagemaker model using boto3 api"""
    template_fields = ('config', 'xcom_tasks', )

    @apply_defaults
    def __init__(self, config, xcom_tasks=None, *args, **kwargs):
        super(CreateModel, self).__init__(*args, **kwargs)
        self.config = config
        self.xcom_tasks = xcom_tasks

    def execute(self, context):
        """
            Execute create model
        """
        import boto3
        from botocore.exceptions import ClientError as botocore_clienterror
        # init boto session and sagemaker
        BOTO_SESSION = boto3.Session()
        SM_CLIENT = BOTO_SESSION.client('sagemaker')

        # context = context_dict + op_kwargs_dict
        training_job_name = self.config.get('training_job_name', None) # pass from training job
        container_image = self.config.get('container_image', None)

        try:
            if training_job_name is None:
                training_job_name = context['task_instance'].xcom_pull(
                    task_ids=self.xcom_tasks['training_job_name']['task_id'],
                    key=self.xcom_tasks['training_job_name']['key']
                )
                print("training_job_name = {}".format(training_job_name))
                assert training_job_name is not None, "No training job name found. Either hardcode in config or pass from previous job"

            # get model_name
            model_name = self.config.get('model_name', training_job_name)

            # retrieve training job config
            training_job_config = SM_CLIENT.describe_training_job(
                TrainingJobName=training_job_name
            )

            if container_image is None:
                container_image=training_job_config['AlgorithmSpecification']['TrainingImage']

            primary_container = {
                'Image': container_image,
                'Mode': self.config.get('primary_container', {}).get('mode','SingleModel'),
                'ModelDataUrl': training_job_config['ModelArtifacts']['S3ModelArtifacts'],
                'Environment': self.config.get('primary_container', {}).get('environment', {})
            }

            try:
                _ = SM_CLIENT.create_model(
                    ModelName=model_name,
                    ExecutionRoleArn=training_job_config['RoleArn'],
                    PrimaryContainer=primary_container,
                    Tags=self.config.get('tags', [])
                )

            except botocore_clienterror as exception:
                if exception.response['Error']['Code'] == 'ValidationException':
                    # handle no
                    if 'Cannot create already existing model' in exception.response['Error']['Message']:
                        print((
                            "This Model already exists, "
                            "with name = {}. Using the already created model."
                            .format(model_name)
                        ))
                    else:
                        raise botocore_clienterror(
                            error_response=exception.operation_name,
                            operation_name=exception.response
                        )

            context['task_instance'].xcom_push(
                key='model_name',
                value=model_name
            )
        except Exception as exception:
            logger.error(exception)

class CreateTransformJob(BaseOperator):
    """Execute Create Sagemaker transform job using boto3 api"""
    template_fields = ('config', 'xcom_tasks', )

    @apply_defaults
    def __init__(self, config, xcom_tasks=None, *args, **kwargs):
        super(CreateTransformJob, self).__init__(*args, **kwargs)
        self.config = config
        self.xcom_tasks = xcom_tasks

    def execute(self, context):
        """
            execute sagemaker transform job
        """
        # init boto session and sagemaker
        BOTO_SESSION = boto3.Session()
        SM_SESSION = Session(default_bucket=self.config.pop('bucket', None), boto_session=BOTO_SESSION)

        model_name = self.config.pop('model_name', None)
        data = self.config.pop('data', None)
        data_type = self.config.pop('data_type','S3Prefix')
        split_type = self.config.pop('split_type','Line')
        compression_type = self.config.pop('compression_type', None)
        content_type = self.config.pop('content_type', 'text/csv')
        input_filter = self.config.pop('input_filter',None)
        join_source = self.config.pop('join_source',None)
        output_filter = self.config.pop('output_filter', None)
        wait_for_completion = self.config.pop('wait', True)
        try:
            if model_name is None:
                # assume model_name is passed
                model_name = context['task_instance'].xcom_pull(
                    task_ids=self.xcom_tasks['model_name']['task_id'],
                    key=self.xcom_tasks['model_name']['key']
                )
                print("model_name = {}".format(model_name))
                logger.info("successfully retrieved model_name = {}".format(model_name))
                assert model_name is not None, "No model name found. Either hardcode in config or pass from previous job"
            
            if data is None:
                # assume model_name is passed
                data = context['task_instance'].xcom_pull(task_ids=self.xcom_tasks['data']['task_id'], key=self.xcom_tasks['data']['key'])
                print("data = {}".format(data))
                logger.info("successfully retrieved data path = {}".format(data))
                assert data is not None, "No s3 data path found. Either hardcode in config or pass from previous job"       

            transformer = Transformer(
                model_name=model_name,
                assemble_with=split_type,
                accept=content_type,
                sagemaker_session=SM_SESSION,
                **self.config
            )
            transformer.transform(
                data=data,
                data_type=data_type,
                split_type=split_type,
                compression_type=compression_type,
                content_type=content_type,
                input_filter=input_filter,
                join_source=join_source,
                output_filter=output_filter,
                wait=wait_for_completion
            )

            context['task_instance'].xcom_push(
                key='transform_job_name',
                value=transformer.latest_transform_job.name
            )
        except Exception as exception:
            logger.error(exception)
            raise



# pylint: disable=super-with-arguments
# pylint: disable=redefined-builtin
# pylint: disable=invalid-name

import logging

import boto3
from airflow.sensors.base_sensor_operator import BaseSensorOperator
from airflow.utils.decorators import apply_defaults

logger = logging.getLogger(None)


class MonitorQuery(BaseSensorOperator):
    """
    Monitors the results from the specified query.
    Only the fields requested in the query are returned,
    along with a @ptr field, which is the identifier for the log record.
    """
    template_fields = ('poke_interval', 'timeout', 'soft_fail', 'mode', 'xcom_tasks')

    @apply_defaults
    def __init__(self, poke_interval=60, timeout=60 * 60 * 24 * 7, soft_fail=False, mode='poke',
                 xcom_tasks=None, *args, **kwargs):
        super(MonitorQuery, self).__init__(
            poke_interval=poke_interval,
            timeout=timeout,
            soft_fail=soft_fail,
            mode=mode,
            *args,
            **kwargs
        )
        self.xcom_tasks = xcom_tasks

    def poke(self, context):
        """Monitors CloudWatch Query Process"""

        BOTO_SESSION = boto3.Session()
        CLIENT = BOTO_SESSION.client('logs')

        task_ids = self.xcom_tasks['query_id']['task_id']
        query_id = context['task_instance'].xcom_pull(
            task_ids=task_ids,
            key=self.xcom_tasks['query_id']['key']
        )

        try:
            query_state = CLIENT.get_query_results(
                queryId=query_id
            )['status']

            if query_state == 'Complete':
                # success
                return True
            elif query_state in ['Running', 'Scheduled']:
                # retry state
                return False
            else:
                # failure, up for retry
                raise Exception(
                    "training job status == {JOB_STATE} with reason = {REASON}".format(
                        JOB_STATE=query_state, REASON=query_state
                    )
                )

        except Exception as err:
            raise



def load_multiple_ecs_tasks(config, operator):
    """
    Designed to load an nth amount of ECS tasks in a single dag
    based on the configuration that hosts multiple verticals:
        <vertical name>>:
          solution: ''
          tags:
            - Key: 'developer'
              Value: ''
          ecs:
            task_cluster: ''
            security_groups:
              - ''
            subnets:
              - ''
            task_definition_names:
              - <here is where all tasks are loaded as a list>
    """
    for task_defintiton_name in config['ecs']['task_definition_names']:
        yield operator(
            task_id=f"run_ecs_task_{task_defintiton_name}",
            config={
                'task_definition_name': task_defintiton_name,
                'task_cluster': config['ecs']['task_cluster'],
                'security_groups': config['ecs']['security_groups'],
                'subnets': config['ecs']['subnets']
            }
        )



# pylint: disable=super-with-arguments
# pylint: disable=redefined-builtin
# pylint: disable=invalid-name

"""
    GLUE TASK CODE
        - these are used in pythonOperator steps as a way \
            to flexibly call boto3 and perform any special requirements
"""
import logging
import time
import warnings

import boto3
from airflow.models import BaseOperator
from airflow.utils.decorators import apply_defaults

logger = logging.getLogger(None)


class RunGlueJob(BaseOperator):
    """Runs Glue Job, and can consume variables for the job."""
    template_fields = ('config', 'xcom_tasks')

    @apply_defaults
    def __init__(self, config, xcom_tasks=None, *args, **kwargs):
        super(RunGlueJob, self).__init__(*args, **kwargs)

        self.config = config
        self.xcom_tasks = xcom_tasks

    def execute(self, context):
        """
            :config.job_name: str. the name of the glue job to run.
            :config.job_arguments: dict. the key value arguments send to the job.
            :xcom_push: str. glue_job_run_id.
        """
        CLIENT = boto3.client('glue')

        job_name = self.config.get('job_name', None)
        job_arguments = self.config.get('job_arguments', None)

        if job_name is None:
            raise ValueError('Job name is required for task.')
        if job_arguments is None:
            warnings.warn('Job arguments are empty.')

        try:
            logging.info('Running glue job {name} with envs {envs}.'.format(
                name=job_name,
                envs=job_arguments)
            )

            response = CLIENT.start_job_run(
                JobName=job_name,
                Arguments=job_arguments
            )

            # push job_run_id to xcom
            context['task_instance'].xcom_push(
                key='job_run_id',
                value=response['JobRunId']
            )

            # push job_name to xcom for other glue tasks to reference
            context['task_instance'].xcom_push(
                key='job_name',
                value=job_name
            )

        except Exception as err:
            raise err


class TagGlueJob(BaseOperator):
    """Adds a tag to the resource specified."""
    template_fields = ('config', 'xcom_tasks',)

    @apply_defaults
    def __init__(self, config, xcom_tasks=None, *args, **kwargs):
        super(TagGlueJob, self).__init__(*args, **kwargs)

        self.config = config
        self.xcom_tasks = xcom_tasks

    def execute(self, context):
        """

        """
        CLIENT = boto3.client('glue')

        job_name = self.config.get('job_name', None)
        resource_tags = self.config.get('resource_tags', None)

        if job_name is None:
            # if optional config job_name is None, pull from xcom
            job_name = context['task_instance'].xcom_pull(
                self.xcom_tasks['job_name']['task_id'],
                key=self.xcom_tasks['job_name']['key']
            )

        if resource_tags is None:
            raise ValueError('Resource tags are required for task.')

        try:
            account_id = boto3.client('sts').get_caller_identity().get('Account')
            glue_resource_arn = 'arn:aws:glue:eu-west-1:{account_id}:job/{name}'.format(
                account_id=account_id, name=job_name
            )

            CLIENT.tag_resource(
                ResourceArn=glue_resource_arn,
                TagsToAdd=resource_tags
            )

            # push resource_tags to xcom for other glue tasks to reference
            context['task_instance'].xcom_push(
                key='resource_tags',
                value=resource_tags
            )

        except Exception as err:
            raise err


class UnTagGlueJob(BaseOperator):
    """Removes a tag to the resource specified."""
    template_fields = ('config', 'xcom_tasks',)

    @apply_defaults
    def __init__(self, config, xcom_tasks=None, *args, **kwargs):
        super(UnTagGlueJob, self).__init__(*args, **kwargs)

        self.config = config
        self.xcom_tasks = xcom_tasks

    def execute(self, context):
        """
            :glue_resource_arn: str. the Arn of the glue resource.
            :resource_tags: dict. the key value pairs of the
                                 relevant tags to remove.
        """
        CLIENT = boto3.client('glue')

        job_name = self.config.get('job_name', None)
        resource_tags = self.config.get('resource_tags', None)

        task_ids = self.xcom_tasks['job_name']['task_id']

        if job_name is None:
            # if optional config job_name is None, pull from xcom
            job_name = context['task_instance'].xcom_pull(
                task_ids=task_ids,
                key=self.xcom_tasks['job_name']['key']
            )

        if resource_tags is None:
            # if optional config resource_tags None, pull from xcom
            resource_tags = context['task_instance'].xcom_pull(
                task_ids=task_ids,
                key=self.xcom_tasks['resource_tags']['key']
            )

        try:
            account_id = boto3.client('sts').get_caller_identity().get('Account')
            glue_resource_arn = 'arn:aws:glue:eu-west-1:{account_id}:job/{name}'.format(
                account_id=account_id, name=job_name
            )

            tags_to_remove = list(key for key in resource_tags.keys())

            CLIENT.untag_resource(
                ResourceArn=glue_resource_arn,
                TagsToRemove=tags_to_remove
            )

        except Exception as err:
            raise err


class RunGlueCrawler(BaseOperator):
    """Runs Glue Crawler."""
    template_fields = ('config', 'xcom_tasks',)

    @apply_defaults
    def __init__(self, config=None, xcom_tasks=None, *args, **kwargs):
        super(RunGlueCrawler, self).__init__(*args, **kwargs)

        self.config = config
        self.xcom_tasks = xcom_tasks

    def execute(self, context):
        """
            :glue_crawler_name: str. the name of the crawler to be triggered.
        """
        CLIENT = boto3.client('glue')

        crawler_name = self.config.get('crawler_name', None)
        if crawler_name is None:
            raise ValueError('Crawler name needed for task.')

        try:
            logging.info(
                'Running glue crawler {name}'.format(
                    name=crawler_name
                )
            )

            CLIENT.start_crawler(
                Name=crawler_name
            )

            # push crawler_name to xcom
            context['task_instance'].xcom_push(
                key='crawler_name',
                value=crawler_name
            )

        except Exception as err:
            raise err



"""
    Sagemaker Sensors

    Used to monitor tasks executed on sagemaker.
    In practice any tasks that have a execution time which could be length or uncertain.
"""
import logging
import boto3
from sagemaker.session import Session
from airflow.sensors.base_sensor_operator import BaseSensorOperator
from airflow.utils.decorators import apply_defaults

logger = logging.getLogger(None)

class MonitorTrainingJob(BaseSensorOperator):
    template_fields = ('training_job_name', 'poke_interval', 'timeout', 'soft_fail', 'mode', 'xcom_tasks')

    @apply_defaults
    def __init__(self, training_job_name=None, poke_interval=60, timeout=60 * 60 * 24 * 7, soft_fail=False, mode='poke', xcom_tasks=None, *args, **kwargs):
        super(MonitorTrainingJob, self).__init__(
            poke_interval=poke_interval,
            timeout=timeout,
            soft_fail=soft_fail,
            mode=mode,
            *args,
            **kwargs
        )
        self.training_job_name = training_job_name
        self.xcom_tasks = xcom_tasks
    
    def poke(self, context):
        """Monitors Sagemaker Transform job"""
        # init boto session and sagemaker
        BOTO_SESSION = boto3.Session()
        SAGEMAKER_CLIENT = BOTO_SESSION.client('sagemaker')
        try:
            if self.training_job_name is None:
                # assume model_name is passed
                self.training_job_name = context['task_instance'].xcom_pull(
                    task_ids=self.xcom_tasks['training_job_name']['task_id'],
                    key=self.xcom_tasks['training_job_name']['key']
                )
                logger.info("successfully retrieved training_job_name = {}".format(self.training_job_name))
                assert self.training_job_name is not None, (
                    "No {xcom_key} found in upstream={task_ids} task. "
                    "Either hardcode in config or pass from previous job".format(
                        xcom_key=self.xcom_tasks['training_job_name']['key'],
                        task_ids=self.xcom_tasks['training_job_name']['task_id']
                    )
                )
            response = SAGEMAKER_CLIENT.describe_training_job(
                TrainingJobName=self.training_job_name
            )
            # push response up
            context['task_instance'].xcom_push(
                key='training_job_name',
                value=self.training_job_name
            )

            job_state = response['TrainingJobStatus']
            reason = response.get('FailureReason', None)
            logger.info("Job status = {}".format(job_state))
            if job_state == 'Completed':
                # failure
                return True
            elif job_state in ['InProgress']:
                # retry state
                return False
            else:
                # failure, up for retry
                raise Exception(
                    "training job status == {JOB_STATE} with reason = {REASON}".format(
                        JOB_STATE=job_state, REASON=reason
                    )
                )
        except Exception as err:
            raise

class MonitorTransformJob(BaseSensorOperator):
    template_fields = ('transform_job_name', 'poke_interval', 'timeout', 'soft_fail', 'mode', 'xcom_tasks')

    @apply_defaults
    def __init__(self, transform_job_name=None, poke_interval=60, timeout=60 * 60 * 24 * 7, soft_fail=False, mode='poke', xcom_tasks=None, *args, **kwargs):
        super(MonitorTransformJob, self).__init__(
            poke_interval=poke_interval,
            timeout=timeout,
            soft_fail=soft_fail,
            mode=mode,
            *args,
            **kwargs
        )
        self.transform_job_name = transform_job_name
        self.xcom_tasks = xcom_tasks

    def poke(self, context):
        """Monitors Sagemaker Transform job"""
        # init boto session and sagemaker
        BOTO_SESSION = boto3.Session()
        SAGEMAKER_CLIENT = BOTO_SESSION.client('sagemaker')
        try:
            if self.transform_job_name is None:
                # assume model_name is passed
                self.transform_job_name = context['task_instance'].xcom_pull(
                    task_ids=self.xcom_tasks['transform_job_name']['task_id'],
                    key=self.xcom_tasks['transform_job_name']['key']
                )
                logger.info("successfully retrieved transform_job_name = {}".format(self.transform_job_name))
                assert self.transform_job_name is not None, (
                    "No {xcom_key} found in upstream={task_ids} task. "
                    "Either hardcode in config or pass from previous job".format(
                        xcom_key=self.xcom_tasks['transform_job_name']['key'],
                        task_ids=self.xcom_tasks['transform_job_name']['task_id']
                    )
                )
            response = SAGEMAKER_CLIENT.describe_transform_job(
                TransformJobName=self.transform_job_name
            )
            # push response up
            context['task_instance'].xcom_push(
                key='transform_job_name',
                value=self.transform_job_name
            )

            job_state = response['TransformJobStatus']
            reason = response.get('FailureReason', None)
            logger.info("Job status = {}".format(job_state))
            if job_state == 'Completed':
                return True
            elif job_state in ['InProgress']:
                return False
            else:
                # retry on failure
                raise Exception(
                    "transform job status == {JOB_STATE} with reason = {REASON}".format(
                        JOB_STATE=job_state, REASON=reason
                    )
                )
        except Exception as err:
            raise



"""
    CALLBACKS
        - python functions executed on specific success/failure events
"""
import logging
from sdc_helpers.slack_helper import SlackHelper

logger = logging.getLogger(None)

def generate_slack_message(
    status: str,
    task: str,
    dag: str,
    exec_date: str,
    log_url: str,
    short_description: str = None,
):
    """generates a slack message from provided dag-task attributes"""
    if status == "success":
        symbol = ":large_green_circle:"
        if short_description is None:
            short_description = "Task Succeeded."
        else:
            short_description = short_description
    elif status == "failure":
        symbol = ":red_circle:"
        if short_description is None:
            short_description = "Task Failed."
        else:
            short_description = short_description
    else:
        symbol = ":warning:"
        if short_description is None:
            short_description = "no detail provided."
        else:
            short_description = short_description
    return """
        {symbol} {short_description}
        *Task*: {task}  
        *Dag*: {dag} 
        *Execution Time*: {exec_date}  
        *Log Url*: {log_url}
    """.format(
        symbol=symbol,
        short_description=short_description,
        task=task,
        dag=dag,
        exec_date=exec_date,
        log_url=log_url,
    )

def slack_notification_success(context):
    ti = context['task_instance']
    slack_msg = generate_slack_message(
        status='success',
        task=ti.task_id,
        dag=ti.dag_id,
        exec_date=context.get('execution_date'),
        log_url=ti.log_url,
    )
    slack_helper = SlackHelper()
    slack_helper.send_critical(message=slack_msg)

def slack_notification_failure(context):
    ti = context['task_instance']
    slack_msg = generate_slack_message(
        status='failure',
        task=ti.task_id,
        dag=ti.dag_id,
        exec_date=context.get('execution_date'),
        log_url=ti.log_url,
    )

    slack_helper = SlackHelper()
    slack_helper.send_critical(message=slack_msg)

def slack_notification_retry(context):
    ti = context['task_instance']
    slack_msg = generate_slack_message(
        status='warning',
        task=ti.task_id,
        dag=ti.dag_id,
        exec_date=context.get('execution_date'),
        log_url=ti.log_url,
        short_description="Task up for retry"
    )

    slack_helper = SlackHelper()
    slack_helper.send_critical(message=slack_msg)

def slack_notification_sla(context):
    ti = context['task_instance']
    slack_msg = generate_slack_message(
        status='warning',
        task=ti.task_id,
        dag=ti.dag_id,
        exec_date=context.get('execution_date'),
        log_url=ti.log_url,
        short_description="Failed to meet SLA"
    )

    slack_helper = SlackHelper()
    slack_helper.send_critical(message=slack_msg)



# pylint: disable=super-with-arguments
# pylint: disable=redefined-builtin
# pylint: disable=invalid-name

"""
    ECS TASK CODE
        - these are used in pythonOperator steps as a way \
            to flexibly call boto3 and perform any special requirements
"""
import logging
import time

import boto3
from airflow.models import BaseOperator
from airflow.utils.decorators import apply_defaults

logger = logging.getLogger(None)


class RunEcsTask(BaseOperator):
    """Run ECS Task Definition."""
    template_fields = ('config', 'xcom_tasks',)

    @apply_defaults
    def __init__(self, config=None, xcom_tasks=None, *args, **kwargs):
        super(RunEcsTask, self).__init__(*args, **kwargs)

        self.config = config
        self.xcom_tasks = xcom_tasks
        self.kwargs = kwargs

    def execute(self, context):
        """

        """
        CLIENT = boto3.client('ecs')

        # for multi_load where config is none, configs can be loaded from kwargs
        if self.config is not None:
            task_definition_name = self.config.get('task_definition_name')
            task_cluster = self.config.get('task_cluster')
            security_groups = self.config.get('security_groups')
            subnets = self.config.get('subnets')
        else:
            task_definition_name = self.kwargs.get('task_definition_name')
            task_cluster = self.kwargs.get('task_cluster')
            security_groups = self.kwargs.get('security_groups')
            subnets = self.kwargs.get('subnets')

        try:
            log_message = """
                Running task: {name} with configurations:
                ecs_task_cluster: {cluster}
                security_groups: {security_groups}
                subnets: {subnets}
                """.format(
                name=task_definition_name,
                cluster=task_cluster,
                security_groups=security_groups,
                subnets=subnets
            )
            logging.info(log_message)

            response = CLIENT.run_task(
                cluster=task_cluster,
                group=task_definition_name,
                launchType='FARGATE',
                networkConfiguration={
                    'awsvpcConfiguration': {
                        'subnets': subnets,
                        'securityGroups': security_groups,
                        'assignPublicIp': 'ENABLED'
                    }
                },
                propagateTags='TASK_DEFINITION',
                taskDefinition=task_definition_name
            )

            logging.info(response)

            ecs_task_run_id = response['tasks'][0]['containers'][0]['taskArn'].split('/')[-1]
            context['task_instance'].xcom_push(
                key='task_run_id',
                value=ecs_task_run_id
            )

        except Exception as err:
            raise err


class MonitorEcsTask(BaseOperator):
    """Monitor ECS Task Definition status for Stopped."""
    template_fields = ('config', 'xcom_tasks',)

    @apply_defaults
    def __init__(self, config, xcom_tasks=None, *args, **kwargs):
        super(MonitorEcsTask, self).__init__(*args, **kwargs)

        self.config = config
        self.xcom_tasks = xcom_tasks

    def execute(self, context):
        """

        """
        CLIENT = boto3.client('ecs')

        task_cluster = self.config.get('task_cluster', None)
        task_run_id = self.config.get('task_run_id', None)

        task_ids = self.xcom_tasks['task_run_id']['task_id']

        if task_run_id is None:
            # if optional config task_run_id is None, pull from xcom
            task_run_id = context['task_instance'].xcom_pull(
                task_ids=task_ids,
                key=self.xcom_tasks['task_run_id']['key']
            )

        try:
            logging.info('Looking for ECS Task ID: {id} in Cluster: {cluster}'.format(
                id=task_run_id, cluster=task_cluster)
            )

            response = CLIENT.describe_tasks(
                cluster=task_cluster,
                tasks=[task_run_id]
            )['tasks']

            logging.info('Response Data: {response}'.format(
                response=response
            ))

            while True:
                ecs_task_definition_status = response[0]['containers'][0].get('lastStatus')

                logging.info('Task Status: {status}.'.format(status=ecs_task_definition_status))
                if ecs_task_definition_status == 'STOPPED':
                    return None

                time.sleep(10)

        except Exception as err:
            raise err



"""
    Athena Sensors
    Used for Monitoring tasks run on athena
    In practice any tasks that have a execution time which could be length or uncertain.
"""
import logging
import boto3
from airflow.sensors.base_sensor_operator import BaseSensorOperator
from airflow.utils.decorators import apply_defaults

logger = logging.getLogger(None)

class MonitorQueryExecution(BaseSensorOperator):
    template_fields = ('query_execution_id', 'poke_interval', 'timeout', 'soft_fail', 'mode', 'xcom_tasks')

    @apply_defaults
    def __init__(
        self,
        query_execution_id=None,
        poke_interval=60,
        timeout=60 * 60 * 24 * 7,
        soft_fail=False,
        mode='poke',
        xcom_tasks=None,
        *args, **kwargs
    ):

        super(MonitorQueryExecution, self).__init__(
            poke_interval=poke_interval,
            timeout=timeout,
            soft_fail=soft_fail,
            mode=mode,
            *args,
            **kwargs
        )
        self.query_execution_id = query_execution_id
        self.xcom_tasks = xcom_tasks
    
    def poke(self, context):
        """
            Runs athena query against feature store
            https://sagemaker.readthedocs.io/en/stable/api/\
                prep_data/feature_store.html#sagemaker.feature_store.feature_group.AthenaQuery
        """
        # init boto session and sagemaker
        BOTO_SESSION = boto3.Session()
        ATHENA_CLIENT = BOTO_SESSION.client('athena')
        try:
            if self.query_execution_id is None:
                self.query_execution_id = context['task_instance'].xcom_pull(
                    task_ids=self.xcom_tasks['query_execution_id']['task_id'],
                    key=self.xcom_tasks['query_execution_id']['key']
                )
                assert self.query_execution_id is not None, (
                    "No {xcom_key} found in upstream={task_ids} task. "
                    "Either hardcode in config or pass from previous job".format(
                        xcom_key=self.xcom_tasks['query_execution_id']['key'],
                        task_ids=self.xcom_tasks['query_execution_id']['task_id']
                    )
                )
            logging.info("With query_id = {}\n".format(self.query_execution_id))
            # loop while results completing
            # on complete, do training next
            response = ATHENA_CLIENT.get_query_execution(
                QueryExecutionId=self.query_execution_id
            )
            query_execution_response = response.get('QueryExecution', {})
            job_state = query_execution_response.get('Status', {}).get('State', None)
            reason = query_execution_response.get('Status', {}).get('StateChangeReason', None)
            output_location = query_execution_response.get('ResultConfiguration', {}).get('OutputLocation', None)
            # push response up
            context['task_instance'].xcom_push(
                key='output_location',
                value=output_location
            )
            logger.info("Job status = {}".format(job_state))
            if job_state == 'SUCCEEDED':
                return True
            elif job_state in ['QUEUED', 'RUNNING']:
                # retry on failure
                return False
            else:
                # retry on failure, eventually fail
                raise Exception(
                    "query status == {JOB_STATE} with reason = {REASON}".format(
                        JOB_STATE=job_state, REASON=reason
                    )
                )
                
        except Exception as err:
            raise



"""
    Set up the sdc_helpers package
"""
from setuptools import setup, find_packages

airflow_requires = [
    'PyYaml',
    'dag-factory',
    'PyYaml',
    'boto3>=1.16.63',
    'sagemaker',
    'sdc_helpers',
    'flatten_dict'
]

def get_required(file_name):
    """
        Get Required Filename during setup
    """
    with open(file_name):
        return file_name.read().splitlines()


setup(
    name='sdc-orchestration-helpers',
    packages=find_packages(exclude=("tests")),
    install_requires=[],
    extras_require={
        'airflow': airflow_requires,
    },
    description='A package of orchestration helpers and utilities for sdc orchestration.',
    version='0.3.3',
    url='https://github.com/RingierIMU/sdc-global-orchestration-helpers',
    author='Ringier South Africa',
    author_email='tools@ringier.co.za',
    keywords=['pip', 'helpers', 'airflow', 'orchestration'],
    download_url='https://github.com/RingierIMU/sdc-global-orchestration-helpers/archive/v0.1.0.zip'
)
